/* _dl_map_object -- Map in a shared object's segments from the file.
   (Upper, machine independent, level)

   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "syscall.h"
#include <link.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include "dynamic-link.h"
#include "dl-internal.h"

/* Try to open NAME in one of the directories in DIRPATH.
   Return the fd, or -1.  If successful, fill in *REALNAME
   with the malloc'd full directory name.  */

static int
open_path (const char *name, size_t namelen,
	   const char *dirpath,
	   char **realname,
	   const char *trusted_dirs[])
{
  char *buf;
  const char *p;
  int fd;

  p = dirpath;
  if (p == NULL || *p == '\0')
    {
      __set_errno (ENOENT);
      return -1;
    }

  buf = __alloca (strlen (dirpath) + 1 + namelen);
  do
    {
      size_t buflen;
      size_t this_len;

      dirpath = p;
      p = strpbrk (dirpath, ":;");
      if (p == NULL)
	p = strchr (dirpath, '\0');

      this_len = p - dirpath;

      /* When we run a setuid program we do not accept any directory.  */
      if (__libc_enable_secure)
	{
	  /* All trusted directory must be complete name.  */
	  if (dirpath[0] != '/')
	    continue;

	  /* If we got a list of trusted directories only accept one
	     of these.  */
	  if (trusted_dirs != NULL)
	    {
	      const char **trust = trusted_dirs;

	      while (*trust !=  NULL)
		if (memcmp (dirpath, *trust, this_len) == 0
		    && (*trust)[this_len] == '\0')
		  break;
		else
		  ++trust;

	      /* If directory is not trusted, ignore this directory.  */
	      if (*trust == NULL)
		continue;
	    }
	}

      if (this_len == 0)
	{
	  /* Two adjacent colons, or a colon at the beginning or the end of
	     the path means to search the current directory.  */
	  (void) memcpy (buf, name, namelen);
	  buflen = namelen;
	}
      else
	{
	  /* Construct the pathname to try.  */
	  (void) memcpy (buf, dirpath, this_len);
	  buf[this_len] = '/';
	  (void) memcpy (&buf[this_len + 1], name, namelen);
	  buflen = this_len + 1 + namelen;
	}

      fd = _dl_sys_open (buf, O_RDONLY);
      if (fd != -1)
	{
	  *realname = _dl_malloc (buflen);
	  if (*realname)
	    {
	      memcpy (*realname, buf, buflen);
	      return fd;
	    }
	  else
	    {
	      /* No memory for the name, we certainly won't be able
		 to load and link it.  */
	      _dl_sys_close (fd);
	      return -1;
	    }
	}
      if (errno != ENOENT && errno != EACCES)
	/* The file exists and is readable, but something went wrong.  */
	return -1;
    }
  while (*p++ != '\0');

  return -1;
}

/* Map in the shared object file NAME.  */

struct link_map *
_dl_map_object (struct link_map *loader, const char *name, int type,
		int trace_mode)
{
  int fd;
  char *realname;
  struct link_map *l;

  /* Look for this name among those already loaded.  */
  for (l = _dl_loaded; l; l = l->l_next)
    if (! strcmp (name, l->l_libname) || /* NAME was requested before.  */
	! strcmp (name, l->l_name) || /* NAME was found before.  */
	/* If the requested name matches the soname of a loaded object,
	   use that object.  */
	(l->l_info[DT_SONAME] &&
	 ! strcmp (name, (const char *) (l->l_addr +
					 (VMA)l->l_info[DT_STRTAB] +
					 (VMA)l->l_info[DT_SONAME]))))
      {
	/* The object is already loaded.
	   Just bump its reference count and return it.  */
	++l->l_opencount;
	return l;
      }

  if (strchr (name, '/') == NULL)
    {
      /* Search for NAME in several places.  */

      size_t namelen = strlen (name) + 1;

      inline void trypath (const char *dirpath, const char *trusted[])
	{
	  fd = open_path (name, namelen, dirpath, &realname, trusted);
	}

      fd = -1;

      /* First try the DT_RPATH of the dependent object that caused NAME
	 to be loaded.  Then that object's dependent, and on up.  */
      for (l = loader; fd == -1 && l; l = l->l_loader)
	if (l && l->l_info[DT_RPATH])
	  trypath ((const char *) (l->l_addr +
				   (VMA)l->l_info[DT_STRTAB] +
				   (VMA)l->l_info[DT_RPATH]), NULL);
      /* If dynamically linked, try the DT_RPATH of the executable itself.  */
      l = _dl_loaded;
      if (fd == -1 && l && l->l_type != lt_loaded && l->l_info[DT_RPATH])
	trypath ((const char *) (l->l_addr +
				 (VMA)l->l_info[DT_STRTAB] +
				 (VMA)l->l_info[DT_RPATH]), NULL);
      /* Try an environment variable (unless setuid).  */
      if (fd == -1 && ! __libc_enable_secure)
	{
	  static const char *trusted_dirs[] =
	  {
#include "trusted-dirs.h"
	    NULL
	  };

	  trypath (getenv ("LD_LIBRARY_PATH"), trusted_dirs);
	}
#ifdef USE_CACHE
      if (fd == -1)
	{
	  /* Check the list of libraries in the file /etc/ld.so.cache,
	     for compatibility with Linux's ldconfig program.  */
	  extern const char *_dl_load_cache_lookup (const char *name);
	  const char *cached = _dl_load_cache_lookup (name);
	  if (cached)
	    {
	      fd = _dl_sys_open (cached, O_RDONLY);
	      if (fd != -1)
		{
		  realname = _dl_strdup (cached);
		  if (realname == NULL)
		    {
		      _dl_sys_close (fd);
		      fd = -1;
		    }
		}
	    }
	}
#endif
      /* Finally, try the default path.  */
      if (fd == -1)
	{
	  extern const char *_dl_rpath;	/* Set in rtld.c. */
	  trypath (_dl_rpath, NULL);
	}
    }
  else
    {
      fd = _dl_sys_open (name, O_RDONLY);
      if (fd != -1)
	{
	  realname = _dl_strdup (name);
	  if (realname == NULL)
	    {
	      _dl_sys_close (fd);
	      fd = -1;
	    }
	}
    }

  if (fd == -1)
    {
      if (trace_mode)
	{
	  /* We haven't found an appropriate library.  But since we
	     are only interested in the list of libraries this isn't
	     so severe.  Fake an entry with all the information we
	     have.  */
	  static const long dummy_bucket = 0;

	  /* Enter the new object in the list of loaded objects.  */
	  if ((l = _dl_new_object (name, name, type)) == NULL)
	    _dl_signal_error (ENOMEM, name,
			      "cannot create shared object descriptor");
	  /* We use an opencount of 0 as a sign for the faked entry.  */
	  l->l_opencount = 0;
	  l->l_reserved = 0;
	  l->l_buckets = &dummy_bucket;
	  l->l_nbuckets = 1;
	  l->l_relocated = 1;

	  return l;
	}
      else
	_dl_signal_error (errno, name, "cannot open shared object file");
    }

  return 
     _dl_load_target_shared_library(name, fd, realname, loader, type);
}
