/* Look up a symbol in the loaded objects.
   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "syscall.h"
#include <stddef.h>
#include <string.h>

#include <link.h>
#include "dl-internal.h"

struct sym_val
  {
    VMA a;
    const PE_syment *s;
  };


/* This is the hashing function specified by the ELF ABI;
   there's no reason to change for PEI/Coff */
static inline unsigned
_dl_elf_hash (const char *name)
{
  unsigned long int hash = 0;
  while (*name != '\0')
    {
      unsigned long int hi;
      hash = (hash << 4) + *name++;
      hi = hash & 0xf0000000;
      if (hi != 0)
	{
	  hash ^= hi >> 24;
	  /* The ELF ABI says `hash &= ~hi', but this is equivalent
	     in this case and on some machines one insn instead of two.  */
	  hash ^= hi;
	}
    }
  return hash;
}


/* Inner part of the lookup functions.  */
static inline VMA
do_lookup (const char **undef_name, unsigned long int *hash,
	   const PE_syment **ref, struct sym_val *result,
	   struct link_map *list[], size_t ii, size_t n,
	   const char *reference_name, struct link_map *skip, int flags)
{
  struct link_map *map;

  /* The name of the symbol we're searching for if an alias was involved.
     This can be static because we're searching for only ONE alias name
     at a time, and if we find an alias for an alias, we want to stomp
     on the prior alias, anyway.  This is only used in the case of a
     short name (so we can get the NUL terminator).
     Note... we have to change the hash, too! */
  static char aliasName[SYMNMLEN+1];
  int recycle = 0;
  int i;

/* We come here if we get an alias; we start searching again in this
   scope using the new name. */

recycle:
  for (i=ii; i < n; ++i)
    {
      VMA symtab;
      const char *strtab;
      int symidx;

      map = list[i];

      /* Here come the extra test needed for `_dl_lookup_symbol_skip'.  */
      if (skip != NULL && map == skip)
	continue;

      /* Don't search the executable when resolving a copy reloc.  */
      if (flags & DL_LOOKUP_NOEXEC && map->l_type == lt_executable)
	continue;

      symtab = (map->l_addr + map->l_info[DT_SYMTAB]);
      strtab = (void *) (map->l_addr + map->l_info[DT_STRTAB]);

      /* Search the appropriate hash bucket in this object's symbol table
	 for a definition for the same symbol name.  */
      for (symidx = map->l_buckets[(*hash) % map->l_nbuckets];
	   symidx != 0;
	   symidx = map->l_chain[symidx])
	{
	  const char *pSymName;
	  char tmpName[SYMNMLEN+1];
	  /* Caution: odd-size structures */
	  const PE_syment *sym = (PE_syment *) (symtab + symidx*PE_SYMENT_SIZE);
	  const PE_syment *sym2;
	  const PE_auxent *aux;
	  int tagidx;

	  if (symidx == map->l_chain[symidx])
	     _dl_signal_error(0, *undef_name, "hash loop");

	  /* At this point, earlier versions of this code used to check 
	     for symbols with a value of zero, and discard them.  Since
	     absolute symbols can (and do) have a value of zero, that
	     check is unsafe.

	     For symbols with a non-zero value, it would also check if
	     the symbol in the undefined section, and if 
	     (flags & DL_LOOKUP_NOPLT) != 0 was also true, skip such
	     symbols.  Since, below, we reject all symbols in the
	     undefined section, that test seems unnecessary.

	     It's unclear whether there ever is a case of a symbol with
	     a non-zero value but in the UNDEF section being useful.
	     The test appeared to be trying to distinguish between a
	     PLT reference to a symbol and some other type, but it's
	     not clear from reading this or ld/bfd, exactly what condition
	     was being reflected.  Gnu ld (for ELF) could produce such
	     a situation (not fully intentionally?), when a PLT entry is 
	     generated, and it would distinguish between a symbol that got
	     into the PLT because it was really undefined (in a shared lib)
	     and one that came from another shared lib.  That, at this point,
	     seems a distinction without a difference.  Doing that can't
	     be carried forward to PEI, because a symbol with a non-zero value
	     in the undefined section is presumed to be the size of
	     a (common?) object. */

	  /* skip non-definitional enteries; NT weaks have a section number
	     of zero, so we have to check for that first. */

	  if (sym->n_sclass != C_NT_WEAK
	     && (sym->n_scnum == N_UNDEF ||
		 sym->n_scnum == N_DEBUG))
	    continue;

	  if (sym->n_zeroes == 0) {
	     pSymName = strtab + sym->n_offset;
	  }
	  else {
	     *(PE_nameunion *)(tmpName) = sym->_n;
	     tmpName[SYMNMLEN] = 0;
	     pSymName = tmpName;
	  }

	  /* There's no type info to check, but if there were, it would
	     come here */

	  if (sym != *ref && strcmp (pSymName, *undef_name))
	    /* Not the symbol we are looking for.  */
	    continue;

	  switch (sym->n_sclass)
	    {
	    case C_EXT:
	      /* Global definition.  Just what we need.  */
	      result->s = sym;
	      result->a = map->l_addr;
	      return 1;
#ifdef USE_WEAK
	    case STB_WEAK:
	      /* Weak definition.  Use this value if we don't find
		 another.  */
	      if (! result->s)
		{
		  result->s = sym;
		  result->a = map->l_addr;
		}
	      break;
#endif
	    case C_NT_WEAK:
	      /* Weak alias; we have the alias name.  Get the real one.
	      What happens here is that when we get an alias name, we
	      change our search target to be for the NEW name.  In this
	      loop, we restart the loop with a new hash value.  In for
	      subsequent scopes, we've changed the search target itself. */

	      aux = (PE_auxent *) (symtab + (symidx+1)*PE_SYMENT_SIZE);
	      tagidx = aux->x_sym.x_tagndx;

	      sym2 = (PE_syment *) (symtab + tagidx*PE_SYMENT_SIZE);
	      if (sym2->n_zeroes == 0) {
		 *undef_name = strtab + sym2->n_offset;
	      }
	      else {
		 *(PE_nameunion *)(aliasName) = sym2->_n;
		 aliasName[SYMNMLEN] = 0;
		 *undef_name = aliasName;
	      }
  	      *hash = _dl_elf_hash (*undef_name);
	      /* Yeah... I know, but C just isn't expressive enough
		 to do this neatly any other way. (We need break n.)  */
	      goto recycle;

	    default:
	      /* Local symbols are ignored.  */
	      break;
	    }
	}
    }

  /* We have not found anything until now.  */
  return 0;
}

/* Search loaded objects' symbol tables for a definition of the symbol
   UNDEF_NAME.  FLAGS is a set of flags.  If DL_LOOKUP_NOEXEC is set,
   then don't search the executable for a definition; this used for
   copy relocs.  If DL_LOOKUP_NOPLT is set, then a PLT entry cannot
   satisfy the reference; some different binding must be found.  */

VMA
_dl_lookup_symbol (const char *needed_name, const PE_syment **ref,
		   struct link_map *symbol_scope[],
		   const char *reference_name,
		   int flags)
{
  unsigned long int hash = _dl_elf_hash (needed_name);
  struct sym_val current_value = { 0, NULL };
  struct link_map **scope;

  /* do_lookup can change undef_name, but leave needed_name alone. */
  const char *undef_name = (char *)needed_name;

  /* Search the relevant loaded objects for a definition.  */
  for (scope = symbol_scope; *scope; ++scope)
    if (do_lookup (&undef_name, &hash, ref, &current_value,
		   (*scope)->l_searchlist, 0, (*scope)->l_nsearchlist,
		   reference_name, NULL, flags))
      break;

  if (current_value.s == NULL &&
      /* This is really a USE_WEAK, but this is much easier */
      ((*ref == NULL) || (*ref)->n_sclass != C_NT_WEAK))
    {
      /* We could find no value for a strong reference.  */
      const char msg[] = "undefined symbol: ";
      const size_t len = strlen (needed_name);
      char buf[sizeof msg + len];
      memcpy (buf, msg, sizeof msg - 1);
      memcpy (&buf[sizeof msg - 1], needed_name, len + 1);
      _dl_signal_error (0, reference_name, buf);
    }

  *ref = current_value.s;
  return current_value.a;
}


/* This function is nearly the same as `_dl_lookup_symbol' but it
   skips in the first list all objects until SKIP_MAP is found.  I.e.,
   it only considers objects which were loaded after the described
   object.  If there are more search lists the object described by
   SKIP_MAP is only skipped.  */
VMA
_dl_lookup_symbol_skip (const char *needed_name, const PE_syment **ref,
			struct link_map *symbol_scope[],
			const char *reference_name,
			struct link_map *skip_map,
			int flags)
{
  unsigned long int hash = _dl_elf_hash (needed_name);
  struct sym_val current_value = { 0, NULL };
  struct link_map **scope;
  size_t i;

  /* do_lookup can change undef_name, but leave needed_name alone. */
  const char *undef_name = (char *)needed_name;

  /* Search the relevant loaded objects for a definition.  */
  scope = symbol_scope;
  for (i = 0; (*scope)->l_dupsearchlist[i] != skip_map; ++i)
    _dl_assert (i < (*scope)->l_ndupsearchlist);

  if (! do_lookup (&undef_name, &hash, ref, &current_value,
		   (*scope)->l_dupsearchlist, i, (*scope)->l_ndupsearchlist,
		   reference_name, skip_map, flags))
    while (*++scope)
      if (do_lookup (&undef_name, &hash, ref, &current_value,
		     (*scope)->l_dupsearchlist, 0, (*scope)->l_ndupsearchlist,
		     reference_name, skip_map, flags))
	break;

  *ref = current_value.s;
  return current_value.a;
}


/* Cache the location of MAP's hash table.  */

void
_dl_setup_hash (struct link_map *map)
{
  long *hash = (long *)(map->l_addr + map->l_info[DT_HASH]);
  int nchain;
  map->l_nbuckets = *hash++;
  nchain = *hash++;
  map->l_buckets = hash;
  hash += map->l_nbuckets;
  map->l_chain = hash;
}
