/* Load a PE/COFF sharable library into memory.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */



/* This file contains the really machine-dependet code to load an image
   into memory and  add the symbol table info to the chain. */

#include <stdlib.h>
#include <string.h>

#include "link.h"
#include "syscall.h"
#include "dl-internal.h"
#include "dynamic-link.h"

/*
 * Read one library into memory, mmap it into the correct locations and
 * add the symbol info to the symbol chain.  Perform any relocations that
 * are required.
 */

struct link_map *_dl_load_target_shared_library (
    const char *libname, 
    int fd,
    char *realname,
    struct link_map *loader,
    int l_type)
{
  PE_filehdr *pFileHdr;
  PE_dyn *pDynInfo;
  struct link_map *pInfo;
  struct link_map *l;
  PE_scnhdr *pSecHdr;
  PE_dyn *ld;
  char *status;
  char header[4096];  /* this is "big enough", but it's not exactly known */
  unsigned int libaddr = 0;
  unsigned long Epa;	/* Entry point (discarded, here) */
  PE_opthdr* pOptHdr;
  
  int i;

#define LOSE(s) lose (0, (s))
  void lose (int code, const char *msg)
    {
      (void) _dl_sys_close (fd);
      if (l)
	{
	  /* Remove the stillborn object from the list and free it.  */
	  if (l->l_prev)
	    l->l_prev->l_next = l->l_next;
	  if (l->l_next)
	    l->l_next->l_prev = l->l_prev;
	  _dl_free (l);
	}
      _dl_free (realname);
      _dl_signal_error (code, libname, msg);
    }

  /* Look again to see if the real name matched another already loaded.  */
  for (l = _dl_loaded; l; l = l->l_next)
    if (!_dl_strcmp (realname, l->l_name))
      {
	/* The object is already loaded.
	   Just bump its reference count and return it.  */
	_dl_sys_close (fd);
	_dl_free (realname);
	++l->l_opencount;
	return l;
      }

  i = _dl_sys_read(fd, header, sizeof(header));
  if (i != sizeof(header))
      lose(errno, "unable to read file header");

  _dl_sys_close(fd);

  pFileHdr = (PE_filehdr *) header;
  pOptHdr = (PE_opthdr*)(header + sizeof(PE_filehdr));

  if (pFileHdr->nt_signature != NT_SIGNATURE || pFileHdr->e_magic != DOSMAGIC) {
    LOSE("not an NT PEI file");
  };

  if(pFileHdr->f_magic != IMAGE_FILE_THIS_MACHINE) {
    LOSE("not an executable for " TARGET_TARGET);
  };

  if((pOptHdr->DllCharacteristics & 0x0001) == 0) {
    LOSE("not a shared library");
  };

  pSecHdr = (PE_scnhdr*)(header + sizeof(PE_filehdr) + sizeof(PE_opthdr));

  if(&pSecHdr[pFileHdr->f_nscns] > header+sizeof(header)) {
    LOSE("Too many sections");
  };

  ld = NULL;
  for(i=0;i < pFileHdr->f_nscns; i++){
    /* Find the .dynamic section */
    if(_dl_strncmp(pSecHdr->s_name, ".dynamic", 8) == 0) {
      if (ld)
	 LOSE("more than one dynamic section");
      ld = (PE_dyn *)pSecHdr->s_vaddr;
    }

    pSecHdr++;
  };

  if(!ld) {
    LOSE("no dynamic section");
  }

  /* Enter the new object in the list of loaded objects.  */
  pInfo = _dl_new_object (realname, libname, l_type);
  if (! pInfo)
    lose (ENOMEM, "cannot create shared object descriptor");
  pInfo->l_opencount = 1;
  pInfo->l_loader = loader;
  pInfo->l_ld = ld;

  libaddr = (unsigned int) __sys_load(realname);

  if (libaddr == (unsigned int)-1) 
    {
      LOSE("system cannot load image");
      return NULL;
    }

  pOptHdr = (PE_opthdr*)(libaddr + sizeof(PE_filehdr));

  pInfo->l_addr = libaddr;
  pInfo->l_offs = libaddr - pOptHdr->ImageBase;

  /* Since we're halfway a DLL, snap the DLL links now */
  i = __sys_reloc(libaddr, 0, 0, &Epa);
  if (i != 0) 
    {
      LOSE("system cannot relocate image");
      return NULL;
    }

  /* scan the dynamic info */
  pInfo->l_ld = (PE_dyn *) ((VMA)pInfo->l_ld + (VMA)libaddr);

  get_dynamic_info(pInfo->l_ld, pInfo->l_info);

  if (pInfo->l_info[DT_HASH])
    _dl_setup_hash (pInfo);
  
  return pInfo;
}
