/* dladdr -- Locate the shared object symbol nearest a given address.
   Copyright (C) 1996 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <stddef.h>
#include <link.h>
#include <dlfcn.h>

#include "syscall.h"

int
dladdr (void *address, Dl_info *info)
{
  const VMA addr = (VMA) address;
  struct link_map *l, *match;
  const PE_syment *matchsym;
  VMA symtab;
  const char *strtab;

  /* Find the highest-addressed object that ADDRESS is not below.  */
  match = NULL;
  for (l = _dl_loaded; l; l = l->l_next)
    if (addr >= l->l_addr && (!match || match->l_addr < l->l_addr))
      match = l;

  if (match)
    {
      PE_scnhdr *pSecHdr;
      int n;

      pSecHdr = (PE_scnhdr *)(match->l_addr + sizeof(PE_filehdr)
		    + sizeof(PE_opthdr));
      /* We know ADDRESS lies within MATCH if in any shared object.
	 Make sure it isn't past the end of MATCH's segments.  */
      n = ((PE_filehdr *)(match->l_addr))->f_nscns;
      do
	--n;
      while ((pSecHdr[n].s_flags & IMAGE_SCN_LNK_REMOVE) == 0);
      if (addr >= (match->l_addr +
		   pSecHdr[n].s_vaddr + pSecHdr[n].s_size))
	/* Off the end of the highest-addressed shared object.  */
	return 0;
    }
  else
    return 0;

  /* Now we know what object the address lies in.  */
  info->dli_fname = match->l_name;
  info->dli_fbase = (void *) match->l_addr;

  symtab = match->l_addr + match->l_info[DT_SYMTAB];
  strtab = ((void *) match->l_addr + match->l_info[DT_STRTAB]);

  /* We assume that the string table follows the symbol table, because
     there is no way to know the size of the dynamic symbol table!!  */
  matchsym = NULL;
  for ( ; (void *) symtab < (void *) strtab; symtab += PE_SYMENT_SIZE)
    {
      PE_syment *sym = (PE_syment *)symtab;

      /* skip non-definitional enteries */
      if ( sym->n_scnum == N_UNDEF || sym->n_scnum == N_DEBUG)
	  continue;

      /* Only external symbols */
      if (sym->n_sclass != C_EXT 
#ifdef USE_WEAK
	&& sym->n_sclass != C_NT_WEAK
#endif
         )
          continue;

      /* Only entries below addr qualify */
      if (match->l_addr + sym->n_value > addr)
	  continue;

      /* Take the first, and replace it with closer ones as we find them */
      if (!matchsym || sym->n_value > matchsym->n_value)
	    matchsym = sym;

    }

  if (matchsym)
    {
      const char *pSymName;
      static char *tmpName[SYMNMLEN+1];

      if (matchsym->n_zeroes == 0) {
	 pSymName = strtab + matchsym->n_offset;
      }
      else {
	 *(PE_nameunion *)(tmpName) = matchsym->_n;
	 tmpName[SYMNMLEN] = 0;
	 pSymName = (const char *)tmpName;
      }

#ifdef TARGET_UNDERSCORE
      if (pSymName[0] == '_') pSymName++;
#endif

      /* We found a symbol close by.  Fill in its name and exact address.  */
      info->dli_sname = pSymName;
      info->dli_saddr = (void *) (match->l_addr + matchsym->n_value);
    }
  else
    {
      /* No symbol matches.  We return only the containing object.  */
      info->dli_sname = NULL;
      info->dli_saddr = NULL;
    }

  return 1;
}
