/* Run time dynamic linker.
   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "syscall.h"

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>		/* Check if MAP_ANON is defined.  */

#include <link.h>
#include "dl-internal.h"
#include "dynamic-link.h"

/* System-dependent function to read a file's whole contents
   in the most convenient manner available.  */
extern void *_dl_sysdep_read_whole_file (const char *filename,
					 size_t *filesize_ptr,
					 int mmap_prot);

/* Helper function to handle errors while resolving symbols.  */
static void print_unresolved (const char *errstring, const char *objname);


int _dl_argc;
char **_dl_argv;
char **_dl_environ;

/* Set nonzero during loading and initialization of executable and
   libraries, cleared before the executable's entry point runs.  This
   must not be initialized to nonzero, because the unused dynamic
   linker loaded in for libc.so's "ld.so.1" dep will provide the
   definition seen by libc.so's initializer; that value must be zero,
   and will be since that dynamic linker's _dl_start and dl_main will
   never be called.  */

int _dl_starting_up;

struct link_map _dl_rtld_map;   /* for ld.so itself */

/* Now life is peachy; we can do all normal operations.
   On to the real work.  */

void _start (void);

unsigned int _dl_skip_args;	/* Nonzero if we were run directly.  */

extern void _dl_fini(void);

/* The real worker in loading the shared libraries (and when run as
 * a command, the main program.)
 *
 * progLoadAddr is where the program was loaded.
 *
 * if standalone is non-zero:
 *    progLoadAddr is probably 0 (and ignored)
 *    startaddr is returned as the entry point address of the main program
 *    that was loaded from the commandline.  Zero to force errors,
 *    otherwise.
 *
 */

void
_dl_main (VMA progLoadAddr, int standalone, VMA *startaddr)
{
  const PE_filehdr *ph;
  struct link_map *progMap;
  struct link_map *l;
  int lazy;
  enum { normal, list, verify, trace } mode;
  struct link_map **preloads;
  unsigned int npreloads;
  const char *preloadlist;
  size_t file_size;
  char *file;
  int i;
  int nsects;
  PE_scnhdr *pSecHdr;

  *startaddr = 0;

  mode = getenv ("LD_TRACE_LOADED_OBJECTS") != NULL ? trace : normal;

  { uid_t uid, euid;
    gid_t gid, egid;

    uid = _dl_sys_getuid();
    euid = _dl_sys_geteuid();
    gid = _dl_sys_getgid();
    egid = _dl_sys_getegid();

    __libc_enable_secure = uid != euid || gid != egid;
  }

  /* LAZY is determined by the parameters --datadeps and --function-deps
     if we trace the binary.  */
  if (mode == trace)
    lazy = -1;
  else
    lazy = !__libc_enable_secure && *(getenv ("LD_BIND_NOW") ?: "") == '\0';

  /* Set up a flag which tells we are just starting.  */
  _dl_starting_up = 1;

  if (standalone)
    {
      /* Ho ho.  We are not the program interpreter!  We are the program
	 itself!  This means someone ran ld.so as a command.  Well, that
	 might be convenient to do sometimes.  We support it by
	 interpreting the args like this:

	 ld.so [--list|--verify|--data-relocs|--function-relocs] PROGRAM ARGS...

	 The first argument is the name of a file containing an 
	 executable we will load and run with the following arguments.
	 To simplify life here, PROGRAM is searched for using the
	 normal rules for shared objects, rather than $PATH or anything
	 like that.  We just load it and use its entry point.
	 This is an easy way to test a new ld.so before installing it.  

	 --list		list the shared libs that are loaded for PROGRAM 
			(at startup).  ARGS lists code or data locations
			to be printed.
	 --verify	just check things out that PROGRAM is a 
			well formed shared lib
	 --data-relocs	explicitly list all unsatisfied data-relocations
			for PROGRAM
	 --function-relocs ditto: functions, too.
	 <none>		Run the program with those args.

	 */
      if (_dl_argc < 2)
	_dl_sysdep_fatal ("\
Usage: ld.so [--list|--verify] EXECUTABLE-FILE [ARGS-FOR-PROGRAM...]\n\
\n\
You have invoked `ld.so', the helper program for shared library executables.\n\
This program usually lives in `/usr/lib/ld.so', and special directives\n\
in executable files using shared libraries tell the program loader\n\
to load the helper program from this file.  This helper program loads\n\
the shared libraries needed by the program executable, prepares the program\n\
to run, and runs it.  You may invoke this helper program directly from the\n\
command line to load and run an executable file; this is like executing\n\
that file itself, but always uses this helper program from the file you\n\
specified, instead of the helper program file specified in the executable\n\
file you run.  The --list option lists all required shared libraries,
and --verify verifies that the program can be successfully executed.\n\
\n\
This is mostly of use for maintainers to test new versions of this\n\
helper program; chances are you did not intend to run this program.\n",
			  NULL);

      /* Note the place where the dynamic linker actually came from.  */
      _dl_rtld_map.l_name = _dl_argv[0];

      while (_dl_argc > 1)
	if (! strcmp (_dl_argv[1], "--list"))
	  {
	    mode = list;
	    lazy = -1;	/* This means do no dependency analysis.  */

	    ++_dl_skip_args;
	    --_dl_argc;
	    ++_dl_argv;
	  }
	else if (! strcmp (_dl_argv[1], "--verify"))
	  {
	    mode = verify;

	    ++_dl_skip_args;
	    --_dl_argc;
	    ++_dl_argv;
	  }
	else if (! strcmp (_dl_argv[1], "--data-relocs"))
	  {
	    mode = trace;
	    lazy = 1;	/* This means do only data relocation analysis.  */

	    ++_dl_skip_args;
	    --_dl_argc;
	    ++_dl_argv;
	  }
	else if (! strcmp (_dl_argv[1], "--function-relocs"))
	  {
	    mode = trace;
	    lazy = 0;	/* This means do also function relocation analysis.  */

	    ++_dl_skip_args;
	    --_dl_argc;
	    ++_dl_argv;
	  }
	else
	  break;

      ++_dl_skip_args;
      --_dl_argc;
      ++_dl_argv;

      if (_dl_argc < 1)
	_dl_sysdep_fatal ("Filename argument required.\n", NULL);

      if (mode == verify)
	{
	  void doit (void)
	    {
	      progMap = _dl_map_object (NULL, _dl_argv[0], lt_library, 0);
	    }
	  char *err_str = NULL;
	  const char *obj_name __attribute__ ((unused));

	  (void) _dl_catch_error (&err_str, &obj_name, doit);
	  if (err_str != NULL)
	    {
	      _dl_free (err_str);
	      _exit (EXIT_FAILURE);
	    }
	}
      else
	{
	   PE_opthdr *header;

	   progMap = _dl_map_object (NULL, _dl_argv[0], lt_library, 0);
	   header = (PE_opthdr *)(progMap->l_addr + sizeof(PE_filehdr));

	   /* This is a "temporary" kluge.  If we're running a main program
	      (not just list) then if the system decides to put it elsewhere
	      than the default, not all the relocations we need are in place.
	      (We either need a .reloc section or need to change cofflink
	      so that RELATIVE relocations are generated.)  The usual
	      symptom is that a jump into a DLL will die.  Since this
	      usage of ld.so is rare, we can fix it if/when we care. */

	   if (header->ImageBase != progMap->l_addr
	      && mode != list)
	      _dl_sysdep_fatal ("Unexpected relocation of main image.\n", NULL);

	   *startaddr = header->standard.entry + progMap->l_addr;
	}

      progMap->l_name = (char *) "";
    }
  else
    {
      /* Create a link_map for the executable itself.
	 This will be what dlopen on "" returns.  */
      progMap = _dl_new_object ((char *) "", "", lt_executable);
      if (progMap == NULL)
	_dl_sysdep_fatal ("cannot allocate memory for link map", NULL);
      progMap->l_addr = progLoadAddr;
    }

  if (progMap != _dl_loaded)
    {
      /* GDB assumes that the first element on the chain is the
	 link_map for the executable itself, and always skips it.
	 Make sure the first one is indeed that one.  */
      progMap->l_prev->l_next = progMap->l_next;
      if (progMap->l_next)
	progMap->l_next->l_prev = progMap->l_prev;
      progMap->l_prev = NULL;
      progMap->l_next = _dl_loaded;
      _dl_loaded->l_prev = progMap;
      _dl_loaded = progMap;
    }

  pSecHdr = (PE_scnhdr *)(progMap->l_addr + sizeof(PE_filehdr)
		+ sizeof(PE_opthdr));

  nsects = ((PE_filehdr *)(progMap->l_addr))->f_nscns;

  /* Scan the program header table for the dynamic section.  */
  for(i=0; i<nsects; i++, pSecHdr++) 
    {
      /* Find the .dynamic section; we can't use strcmp yet because
         literals aren't relocated yet.  */
      if (pSecHdr->s_name[1] == 'd' &&
          pSecHdr->s_name[2] == 'y' &&
          pSecHdr->s_name[3] == 'n' &&
          pSecHdr->s_name[4] == 'a' &&
          pSecHdr->s_name[5] == 'm' &&
          pSecHdr->s_name[6] == 'i' &&
          pSecHdr->s_name[7] == 'c' &&
          pSecHdr->s_name[0] == '.') 
        {
	  progMap->l_ld = (void *) progMap->l_addr + pSecHdr->s_vaddr;
	  break;
        }
    }

    if (mode == normal && progMap->l_ld == NULL)
      {
         _dl_sysdep_fatal (".dynamic section not found in main program\n",
	     NULL);
	 _exit(0377);
      }

  if (! _dl_rtld_map.l_libname && _dl_rtld_map.l_name)
    /* We were invoked directly, so the program might not have a PT_INTERP.  */
    _dl_rtld_map.l_libname = _dl_rtld_map.l_name;
  else
    _dl_assert (_dl_rtld_map.l_libname); /* How else did we get here?  */

  if (mode == verify)
    /* We were called just to verify that this is a dynamic executable
       using us as the program interpreter.  */
    _exit (progMap->l_ld == NULL ? EXIT_FAILURE : EXIT_SUCCESS);

  /* Extract the contents of the dynamic section for easy access.  */
  get_dynamic_info (progMap->l_ld, progMap->l_info);
  if (progMap->l_info[DT_HASH])
    /* Set up our cache of pointers into the hash table.  */
    _dl_setup_hash (progMap);

  /* Put the link_map for ourselves on the chain so it can be found by
     name.  */
  if (! _dl_rtld_map.l_name)
    /* If not invoked directly, the dynamic linker shared object file was
       found by the PT_INTERP name.  */
    _dl_rtld_map.l_name = (char *) _dl_rtld_map.l_libname;
  _dl_rtld_map.l_type = lt_library;

  /* Starting from the main, find the end, and plug ourselves in there */
  l = progMap;
  while (l->l_next)
    l = l->l_next;
  l->l_next = &_dl_rtld_map;

  _dl_rtld_map.l_prev = l;

  /* We have two ways to specify objects to preload: via environment
     variable and via the file /etc/ld.so.preload.  The later can also
     be used when security is enabled.  */
  preloads = NULL;
  npreloads = 0;

  preloadlist = getenv ("LD_PRELOAD");
  if (preloadlist)
    {
      /* The LD_PRELOAD environment variable gives a white space
	 separated list of libraries that are loaded before the
	 executable's dependencies and prepended to the global scope
	 list.  If the binary is running setuid all elements
	 containing a '/' are ignored since it is insecure.  */
      char *list;
      char *p;

      list = alloca(strlen(preloadlist)+1);
      memcpy(list, preloadlist, strlen(preloadlist)+1);

      while ((p = strsep (&list, " ")) != NULL)
	if (! __libc_enable_secure || strchr (p, '/') == NULL)
	  {
	    (void) _dl_map_object (NULL, p, lt_library, 0);
	    ++npreloads;
	  }
    }

  /* We don't do ld.so.preload at this time */
#if USE_CACHE /* [ */
  /* Read the contents of the file.  */
  file = _dl_sysdep_read_whole_file ("/etc/ld.so.preload", &file_size,
				     PROT_READ | PROT_WRITE);
  if (file)
    {
      /* Parse the file.  It contains names of libraries to be loaded,
	 separated by white spaces or `:'.  It may also contain
	 comments introduced by `#'.  */
      char *problem;
      char *runp;
      size_t rest;

      /* Eliminate comments.  */
      runp = file;
      rest = file_size;
      while (rest > 0)
	{
	  char *comment = memchr (runp, '#', rest);
	  if (comment == NULL)
	    break;

	  rest -= comment - runp;
	  do
	    *comment = ' ';
	  while (--rest > 0 && *++comment != '\n');
	}

      /* We have one problematic case: if we have a name at the end of
	 the file without a trailing terminating characters, we cannot
	 place the \0.  Handle the case separately.  */
      if (file[file_size - 1] != ' ' && file[file_size] != '\t'
	  && file[file_size] != '\n')
	{
	  problem = &file[file_size];
	  while (problem > file && problem[-1] != ' ' && problem[-1] != '\t'
		 && problem[-1] != '\n')
	    --problem;

	  if (problem > file)
	    problem[-1] = '\0';
	}
      else
	problem = NULL;

      if (file != problem)
	{
	  char *p;
	  runp = file;
	  while ((p = strsep (&runp, ": \t\n")) != NULL)
	    {
	      (void) _dl_map_object (NULL, p, lt_library, 0);
	      ++npreloads;
	    }
	}

      if (problem != NULL)
	{
	  char *p = strndupa (problem, file_size - (problem - file));
	  (void) _dl_map_object (NULL, p, lt_library, 0);
	}

      /* We don't need the file anymore.  */
      __munmap (file, file_size);
    }
#endif /* ] */

  if (npreloads != 0)
    {
      /* Set up PRELOADS with a vector of the preloaded libraries.  */
      unsigned int i;
      preloads = __alloca (npreloads * sizeof preloads[0]);
      l = _dl_rtld_map.l_next; /* End of the chain before preloads.  */
      i = 0;
      do
	{
	  preloads[i++] = l;
	  l = l->l_next;
	} while (l);
      _dl_assert (i == npreloads);
    }

  /* Load all the libraries specified by DT_NEEDED entries.  If LD_PRELOAD
     specified some libraries to load, these are inserted before the actual
     dependencies in the executable's searchlist for symbol resolution.  */
  _dl_map_object_deps (progMap, preloads, npreloads, mode == trace);

#ifndef MAP_ANON
  /* We are done mapping things, so close the zero-fill descriptor.  */
  __close (_dl_zerofd);
  _dl_zerofd = -1;
#endif

  /* Remove _dl_rtld_map from the chain.  */
  _dl_rtld_map.l_prev->l_next = _dl_rtld_map.l_next;
  if (_dl_rtld_map.l_next)
    _dl_rtld_map.l_next->l_prev = _dl_rtld_map.l_prev;

  if (_dl_rtld_map.l_opencount)
    {
      /* Some DT_NEEDED entry referred to the interpreter object itself, so
	 put it back in the list of visible objects.  We insert it into the
	 chain in symbol search order because gdb uses the chain's order as
	 its symbol search order.  */
      unsigned int i = 1;
      while (progMap->l_searchlist[i] != &_dl_rtld_map)
	++i;
      _dl_rtld_map.l_prev = progMap->l_searchlist[i - 1];
      _dl_rtld_map.l_next = (i + 1 < progMap->l_nsearchlist ?
			     progMap->l_searchlist[i + 1] : NULL);
      _dl_assert (_dl_rtld_map.l_prev->l_next == _dl_rtld_map.l_next);
      _dl_rtld_map.l_prev->l_next = &_dl_rtld_map;
      if (_dl_rtld_map.l_next)
	{
	  _dl_assert (_dl_rtld_map.l_next->l_prev == _dl_rtld_map.l_prev);
	  _dl_rtld_map.l_next->l_prev = &_dl_rtld_map;
	}
    }

  if (mode != normal)
    {
      /* We were run just to list the shared libraries.  It is
	 important that we do this before real relocation, because the
	 functions we call below for output may no longer work properly
	 after relocation.  */

      int i;

      if (! _dl_loaded->l_info[DT_NEEDED])
	_dl_sysdep_message ("\t", "statically linked\n", NULL);
      else
	for (progMap = _dl_loaded->l_next; progMap; progMap = progMap->l_next)
	  if (progMap->l_opencount == 0)
	    /* The library was not found.  */
	    _dl_sysdep_message ("\t", progMap->l_libname, " => not found\n", NULL);
	  else
	    {
	      char buf[20], *bp;
	      buf[sizeof buf - 1] = '\0';
	      bp = _itoa (progMap->l_addr, &buf[sizeof buf - 1], 16, 0);
	      while ((size_t) (&buf[sizeof buf - 1] - bp)
		     < sizeof progMap->l_addr * 2)
		*--bp = '0';
	      _dl_sysdep_message ("\t", progMap->l_libname, " => ", progMap->l_name,
				  " (0x", bp, ")\n", NULL);
	    }

      if (mode != trace)
	for (i = 1; i < _dl_argc; ++i)
	  {
	    const PE_syment *ref = NULL;
	    int loadbase = _dl_lookup_symbol (_dl_argv[i], &ref,
						     &_dl_default_scope[2],
						     "argument",
						     DL_LOOKUP_NOPLT);
	    char buf[20], *bp;
	    buf[sizeof buf - 1] = '\0';
	    bp = _itoa (ref->n_value, &buf[sizeof buf - 1], 16, 0);
	    while ((size_t) (&buf[sizeof buf - 1] - bp) < sizeof loadbase * 2)
	      *--bp = '0';
	    _dl_sysdep_message (_dl_argv[i], " found at 0x", bp, NULL);
	    buf[sizeof buf - 1] = '\0';
	    bp = _itoa (loadbase, &buf[sizeof buf - 1], 16, 0);
	    while ((size_t) (&buf[sizeof buf - 1] - bp) < sizeof loadbase * 2)
	      *--bp = '0';
	    _dl_sysdep_message (" in object at 0x", bp, "\n", NULL);
	  }
      else if (lazy >= 0)
	{
	  /* We have to do symbol dependency testing.  */
	  void doit (void)
	    {
	      _dl_relocate_object (progMap, _dl_object_relocation_scope (progMap), lazy);
	    }

	  progMap = _dl_loaded;
	  while (progMap->l_next)
	    progMap = progMap->l_next;
	  do
	    {
	      if (progMap != &_dl_rtld_map && progMap->l_opencount > 0)
		{
		  _dl_receive_error (print_unresolved, doit);
		  *_dl_global_scope_end = NULL;
		}
	      progMap = progMap->l_prev;
	    } while (progMap);
	}

      _exit (0);
    }

  {
    /* Now we have all the objects loaded.  Relocate them all except for
       the dynamic linker itself.  We do this in reverse order so that copy
       relocs of earlier objects overwrite the data written by later
       objects.  We do not re-relocate the dynamic linker itself in this
       loop because that could result in the GOT entries for functions we
       call being changed, and that would break us.  It is safe to relocate
       the dynamic linker out of order because it has no copy relocs (we
       know that because it is self-contained).  */

    l = _dl_loaded;
    while (l->l_next)
      l = l->l_next;
    do
      {
	if (l != &_dl_rtld_map)
	  {
	    _dl_relocate_object (l, _dl_object_relocation_scope (l), lazy);
	    *_dl_global_scope_end = NULL;
	  }
	l = l->l_prev;
      } while (l);

    if (_dl_rtld_map.l_opencount > 0)
      /* There was an explicit ref to the dynamic linker as a shared lib.
	 Re-relocate ourselves with user-controlled symbol definitions.  */
      _dl_relocate_object (&_dl_rtld_map, &_dl_default_scope[2], 0);
  }

  {
    /* Initialize _r_debug.  */
    struct r_debug *r = _dl_debug_initialize (_dl_rtld_map.l_addr);

    l = _dl_loaded;

    if (l->l_info[DT_DEBUG]) 
      {
      /* There is a DT_DEBUG entry in the dynamic section.  Fill it in
	 with the run-time address of the r_debug structure; gotta
	 deal with protection. */
	VMA addr;
	addr = (VMA)(l->l_info[DT_DEBUG]);
	_dl_sys_mprotect((void *)(addr & ~(PAGESZ-1)),PAGESZ,
	    PROT_READ|PROT_WRITE);
	*(VMA*)addr = (VMA)r;
	_dl_sys_mprotect((void *)(addr & ~(PAGESZ-1)),PAGESZ,
	    PROT_READ);
      }

    /* Fill in the pointer in the dynamic linker's own dynamic section, in
       case you run gdb on the dynamic linker directly.  */
    if (_dl_rtld_map.l_info[DT_DEBUG])
      {
	VMA addr;
	addr = (VMA)(_dl_rtld_map.l_info[DT_DEBUG]);
	_dl_sys_mprotect((void *)(addr & ~(PAGESZ-1)),PAGESZ,
	    PROT_READ|PROT_WRITE);
	*(VMA*)addr = (VMA)r;
	_dl_sys_mprotect((void *)(addr & ~(PAGESZ-1)),PAGESZ,
	    PROT_READ);
      }

    /* Notify the debugger that all objects are now mapped in.  */
    r->r_state = RT_ADD;
    _dl_debug_state ();
   
  }

  {
    VMA init;
    /* Run the initializer functions of new objects.  */
    while (init = _dl_init_next (_dl_default_scope[2]))
      (*(void (*) (int, char **, char **)) init) 
	  (_dl_argc, _dl_argv, _dl_environ);
  }

  _dl_starting_up = 0;
}

/* This is a little helper function for resolving symbols while
   tracing the binary.  */
static void
print_unresolved (const char *errstring, const char *objname)
{
  _dl_sysdep_error (errstring, "	(", objname, ")\n", NULL);
}
